
/* GTK */
#include <glib.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <string.h>
#include <math.h>
#include <stdio.h>

#include "unicode_names.h"
#include "unicode_categories.h"
#include "unicode_tools.c"

// global widgets, et cetera...
gchar *selection, *selection_source;
GtkWidget *notebook;
GtkWidget *list;	// tab nr. 1 : list
GtkWidget *tview;	// tab nr. 2 : text
GtkWidget *iview;	// tab nr. 3 : info...
GtkToolItem *button_index;
GtkWidget *statusbar;
guint statusbar_context;
GtkWidget * toolbar_entry;

// prototype // TODO -> *.h
void info_show_codepoint (GtkTextBuffer *buffer, gunichar ch);
void set_statusbar ();

/*
 * SIGNALS  -----------------------------------------------------------------
 *
 */


gboolean on_key_press (GtkWidget * window, GdkEventKey*	pKey, gpointer userdata){
  if (pKey->type == GDK_KEY_PRESS){

    // DEBUG printf ("[KEY:%s|%x|%x]\n", pKey->string, pKey->hardware_keycode, pKey->state);

    // close on ESC...
    if (pKey->string[0]==0x1b) gtk_main_quit();
  }
  return FALSE;
}

/*
  // Ask to receive events the drawing area doesn't normally subscribe to
  gtk_widget_set_events (darea, gtk_widget_get_events (darea)
                                | GDK_LEAVE_NOTIFY_MASK
                                | GDK_BUTTON_PRESS_MASK
                                | GDK_POINTER_MOTION_MASK
                                | GDK_POINTER_MOTION_HINT_MASK);
*/

/*
 * LIST  -----------------------------------------------------------------
 *
 */

enum {
  LVIEW_ID=0,
  LVIEW_CODE,
  LVIEW_CHAR,
  LVIEW_ENTITY,
  LVIEW_CODEPOINT,
  LVIEW_NAME,
  LVIEW_CATEGORY,
  LVIEW_CATEGORY_COLOR,
  LVIEW_NCOLS,
};

// TODO TODO: barvy, unicode kategorie, ....

GtkListStore *make_model (gchar *sel) {

  // model, data, ...
  GtkListStore *store;
  GtkTreeIter iter;

  // widgets ....
  GtkWidget *cbox;
  GtkCellRenderer *cell;

  // 1. model...
  store = gtk_list_store_new( LVIEW_NCOLS,
             G_TYPE_UINT,   // LVIEW_ID
             G_TYPE_UINT,   // LVIEW_CODE
             G_TYPE_STRING, // LVIEW_CHAR
             G_TYPE_STRING, // LVIEW_ENTITY
             G_TYPE_STRING, // LVIEW_CODEPOINT
             G_TYPE_STRING, // LVIEW_NAME
             G_TYPE_STRING, // LVIEW_CATEGORY
             G_TYPE_STRING  // LVIEW_CATEGORY_COLOR
             ); 


  // 2. fill in the data...
  gchar *s=sel, *chs;
  gunichar ch;
  guint i=0;

  if (s && strlen(s)) while (ch=g_utf8_get_char_validated(s, -1)) {

    chs=g_malloc0(8);
    g_unichar_to_utf8(ch, chs);

    // DEBUG
    //printf ("\"%s\"\tU+%04X\t%s\n", chs, ch, name_for_codepoint(ch));

    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter, LVIEW_ID, i,
                                     LVIEW_CODE, ch,
                                     LVIEW_CHAR, chs,
                                     LVIEW_ENTITY, g_strdup_printf("&#%d;", ch), 
                                     LVIEW_CODEPOINT, g_strdup_printf("U+%04X", ch), 
                                     LVIEW_NAME, name_for_codepoint(ch),
                                     LVIEW_CATEGORY, unicode_categories[g_unichar_type(ch)],
                                     LVIEW_CATEGORY_COLOR, unicode_categories_colors[g_unichar_type(ch)],
                                     -1); 

    s=g_utf8_next_char(s); i++;
  }

  return store;
}

void list_row_activated (GtkTreeView *treeview, GtkTreePath *path, GtkTreeViewColumn *col, gpointer userdata)
{
  GtkTreeModel *model;
  GtkTreeIter   iter;
  
  model = gtk_tree_view_get_model(treeview);
  
  if (gtk_tree_model_get_iter(model, &iter, path))
  {
    gunichar ch;
    gtk_tree_model_get(model, &iter, LVIEW_CODE, &ch, -1);
    info_show_codepoint (gtk_text_view_get_buffer (GTK_TEXT_VIEW (iview)), ch);
    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook), 1);
  }
}


static void list_init (GtkWidget *list)
{

  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes("Char", renderer,
                                                    "text", LVIEW_CHAR,
                                                    "foreground", LVIEW_CATEGORY_COLOR,
                                                    NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes("Codepoint", renderer,
                                                    "text", LVIEW_CODEPOINT,
                                                    "foreground", LVIEW_CATEGORY_COLOR,
                                                    NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes("Char Name", renderer,
                                                    "text", LVIEW_NAME,
                                                    "foreground", LVIEW_CATEGORY_COLOR,
                                                    NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);

  renderer = gtk_cell_renderer_text_new();
  // TODO: viz http://www.w3schools.com/tags/ref_entities.asp ...
  column = gtk_tree_view_column_new_with_attributes("XML Entity", renderer,
                                                    "text", LVIEW_ENTITY,
                                                    "foreground", LVIEW_CATEGORY_COLOR,
                                                    NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);

  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes("Category", renderer,
                                                    "text", LVIEW_CATEGORY,
                                                    "foreground", LVIEW_CATEGORY_COLOR,
                                                    NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(list), column);

  g_signal_connect(list, "row-activated", G_CALLBACK (list_row_activated), NULL);
}

/*
 * INFO  -----------------------------------------------------------------
 *
 */

void insert_codepoint_link (GtkTextBuffer *buffer, GtkTextIter *iter, gunichar ch) {
  GtkTextTag *tag;

  tag = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "foreground", "blue", 
                                    "underline", PANGO_UNDERLINE_SINGLE, 
                                    NULL);
  g_object_set_data (G_OBJECT (tag), "link_type", "codepoint");
  g_object_set_data (G_OBJECT (tag), "codepoint", GUINT_TO_POINTER (ch));
  gtk_text_buffer_insert_with_tags (buffer, iter,
                                    g_strdup_printf("U+%04X\t%s", ch, name_for_codepoint(ch)), -1,
                                    tag, NULL);
}

void insert_index_link (GtkTextBuffer *buffer, GtkTextIter *iter, gchar *text) {
  GtkTextTag *tag;

  tag = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "foreground", "blue", 
                                    "underline", PANGO_UNDERLINE_SINGLE, 
                                    NULL);
  g_object_set_data (G_OBJECT (tag), "link_type", "index");
  gtk_text_buffer_insert_with_tags (buffer, iter, text, -1, tag, NULL);
}

void insert_url_link (GtkTextBuffer *buffer, GtkTextIter *iter, gchar *text, gchar *url) {
  GtkTextTag *tag;

  tag = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "foreground", "blue", 
                                    "underline", PANGO_UNDERLINE_SINGLE, 
                                    NULL);
  g_object_set_data (G_OBJECT (tag), "link_type", "url");
  g_object_set_data (G_OBJECT (tag), "url", url);
  gtk_text_buffer_insert_with_tags (buffer, iter, text, -1, tag, NULL);
}

void info_show_overview (GtkTextBuffer *buffer, gchar *selection) {
  GtkTextIter	iter;
  GtkTextTag	*chr;

  chr = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "foreground", "black", 
                                    "background", "grey",
                                    NULL);

  gtk_text_buffer_set_text (buffer, "", 0);
  gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);

  // 2. fill in the data...
  gchar *s=selection, *chs;
  gunichar ch;

  if (s && strlen(s)) while (ch=g_utf8_get_char_validated(s, -1)) {

    chs=g_malloc0(8);
    g_unichar_to_utf8(ch, chs);

    gtk_text_buffer_insert_with_tags (buffer, &iter, g_strdup_printf("\t%s\t", chs, ch), -1, chr, NULL);
    gtk_text_buffer_insert (buffer, &iter, "\t", -1);
    insert_codepoint_link (buffer, &iter, ch);
    gtk_text_buffer_insert (buffer, &iter, "\n", -1);

    // TODO: free chs
    s=g_utf8_next_char(s);
  }

  // now we are in index - make the button insensitive...
  gtk_widget_set_sensitive(GTK_WIDGET(button_index), FALSE);
  gtk_widget_set_sensitive(GTK_WIDGET(toolbar_entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(toolbar_entry), "");
}

void info_show_codepoint (GtkTextBuffer *buffer, gunichar ch) {
  GtkTextIter    iter;
  GtkTextTag *hdr, *chr, *bold, *ctg;
  gchar *chs;

  hdr = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "weight", PANGO_WEIGHT_BOLD, 
                                    "size", 22 * PANGO_SCALE, 
                                    "justification", GTK_JUSTIFY_CENTER,
                                    NULL);

  chr = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "foreground", "white", 
                                    "background", "green",
                                    "justification", GTK_JUSTIFY_CENTER,
                                    "size", 70 * PANGO_SCALE, 
                                    NULL);

  bold = gtk_text_buffer_create_tag  (buffer, NULL, 
                                    "weight", PANGO_WEIGHT_BOLD, 
                                    NULL);

  gtk_text_buffer_set_text (buffer, "", 0);
  gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);

  chs=g_malloc0(8);
  g_unichar_to_utf8(ch, chs);

  // a) heading
  gtk_text_buffer_insert_with_tags (buffer, &iter,
                                    g_strdup_printf("U+%04X %s\n", ch, name_for_codepoint(ch)), -1,
                                    hdr, NULL);
  // b) the character itself
  gtk_text_buffer_insert_with_tags (buffer, &iter, chs, -1, chr, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\n\n\n", -1);

  // c) codepoint
  gtk_text_buffer_insert_with_tags (buffer, &iter, "Codepoint:", -1, bold, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\t", -1);
  gtk_text_buffer_insert (buffer, &iter, g_strdup_printf("U+%04X\n", ch), -1);

  // d) unicode name
  gtk_text_buffer_insert_with_tags (buffer, &iter, "Unicode Name:", -1, bold, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\t", -1);
  gtk_text_buffer_insert (buffer, &iter, name_for_codepoint(ch), -1);
  gtk_text_buffer_insert (buffer, &iter, "\n", -1);

  // e) XML entity
  // TODO: http://www.w3schools.com/tags/ref_entities.asp
  // TODO: http://www.w3schools.com/html/html_entities.asp
  gtk_text_buffer_insert_with_tags (buffer, &iter, "XML Entity:", -1, bold, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\t", -1);
  gtk_text_buffer_insert (buffer, &iter, g_strdup_printf("&#%d;\n", ch), -1);

  // f) category
  ctg = gtk_text_buffer_create_tag (buffer, NULL, 
                                    "foreground", unicode_categories_colors[g_unichar_type(ch)],
                                    NULL);
  gtk_text_buffer_insert_with_tags (buffer, &iter, "Category:", -1, bold, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\t\t", -1);
  gtk_text_buffer_insert_with_tags (buffer, &iter, unicode_categories[g_unichar_type(ch)], -1, ctg, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\n", -1);

  // ---------- online: 
  gtk_text_buffer_insert_with_tags (buffer, &iter, "Online Info:", -1, bold, NULL);
  gtk_text_buffer_insert (buffer, &iter, "\t", -1);
  insert_url_link (buffer, &iter, "fileformat.info",
         g_strdup_printf("http://www.fileformat.info/info/unicode/char/%x/index.htm", ch));

  // TODO TODO TODO: à la: http://www.charbase.com/0074-unicode-latin-small-letter-t 
  gtk_text_buffer_insert (buffer, &iter, "\n", -1);

  gtk_text_buffer_insert (buffer, &iter, "\n\n[«", -1);
  insert_index_link (buffer, &iter, "back");
  gtk_text_buffer_insert (buffer, &iter, "]\n", -1);


  // allow returning to the INDEX...
  gtk_widget_set_sensitive(GTK_WIDGET(button_index), TRUE);
  gtk_entry_set_text (GTK_ENTRY(toolbar_entry), g_strdup_printf("U+%04X %s", ch, name_for_codepoint(ch)));
  gtk_widget_set_sensitive(GTK_WIDGET(toolbar_entry), TRUE);
}

static void follow_if_link (GtkWidget   *text_view, GtkTextIter *iter)
{
  GSList *tags = NULL, *tagp = NULL;
  gchar *link_type;
  gunichar codepoint;
  gchar *link, *s;

  tags = gtk_text_iter_get_tags (iter);
  for (tagp = tags;  tagp != NULL;  tagp = tagp->next)
  {
    GtkTextTag *tag = tagp->data;
    link_type = g_object_get_data (G_OBJECT (tag), "link_type");
    if (link_type && !strcmp(link_type, "codepoint"))
    {
      codepoint = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "codepoint"));

      if (codepoint)
      {
        info_show_codepoint (gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view)), codepoint);
        break;
      }
    }
    else if (link_type && !strcmp(link_type, "url"))
    {
      link = g_object_get_data (G_OBJECT (tag), "url");
      s = g_strdup_printf("/usr/bin/gnome-open '%s'", link);
      g_spawn_command_line_async(s, NULL);
      g_free(s);
    }
    else if (link_type && !strcmp(link_type, "index"))
    {
      info_show_overview (gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view)), selection);
      break;
    }
  }

  if (tags) 
    g_slist_free (tags);
}

static gboolean info_key_pressed (GtkWidget *text_view, GdkEventKey *event)
{
  GtkTextIter iter;
  GtkTextBuffer *buffer;

  switch (event->keyval)
    {
      case GDK_Return: 
      case GDK_KP_Enter:
        buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view));
        gtk_text_buffer_get_iter_at_mark (buffer, &iter, 
                                          gtk_text_buffer_get_insert (buffer));
        follow_if_link (text_view, &iter);
        break;

      default:
        break;
    }

  return FALSE;
}

/* Click...
 *  */
static gboolean info_event_after (GtkWidget *text_view, GdkEvent  *ev)
{
  GtkTextIter start, end, iter;
  GtkTextBuffer *buffer;
  GdkEventButton *event;
  gint x, y;

  if (ev->type != GDK_BUTTON_RELEASE)
    return FALSE;

  event = (GdkEventButton *)ev;

  if (event->button != 1)
    return FALSE;

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view));

  /* we shouldn't follow a link if the user has selected something */
  gtk_text_buffer_get_selection_bounds (buffer, &start, &end);
  if (gtk_text_iter_get_offset (&start) != gtk_text_iter_get_offset (&end))
    return FALSE;

  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (text_view), 
                                         GTK_TEXT_WINDOW_WIDGET,
                                         event->x, event->y, &x, &y);

  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (text_view), &iter, x, y);

  follow_if_link (text_view, &iter);

  return FALSE;
}

static gboolean hovering_over_link = FALSE;
static GdkCursor *hand_cursor = NULL;
static GdkCursor *regular_cursor = NULL;

static void set_cursor_if_appropriate (GtkTextView *text_view, gint x, gint y)
{
  GSList *tags = NULL, *tagp = NULL;
  GtkTextIter iter;
  gboolean hovering = FALSE;
  gchar *link_type;

  gtk_text_view_get_iter_at_location (text_view, &iter, x, y);


  tags = gtk_text_iter_get_tags (&iter);
  for (tagp = tags;  tagp != NULL;  tagp = tagp->next)
    {
      GtkTextTag *tag = tagp->data;
      link_type = g_object_get_data (G_OBJECT (tag), "link_type");

      if (link_type && strcmp(link_type, "")) 
        {
          hovering = TRUE;
          break;
        }
    }

  if (hovering != hovering_over_link)
    {
      hovering_over_link = hovering;

      if (hovering_over_link)
        gdk_window_set_cursor (gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT), hand_cursor);
      else
        gdk_window_set_cursor (gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT), regular_cursor);
    }

  if (tags) 
    g_slist_free (tags);
}

static gboolean info_motion_notify_event (GtkWidget *text_view, GdkEventMotion *event)
{
  gint x, y;

  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (text_view), 
                                         GTK_TEXT_WINDOW_WIDGET,
                                         event->x, event->y, &x, &y);

  set_cursor_if_appropriate (GTK_TEXT_VIEW (text_view), x, y);

  gdk_window_get_pointer (text_view->window, NULL, NULL, NULL);
  return FALSE;
}

static gboolean info_visibility_notify_event (GtkWidget *text_view, GdkEventVisibility *event)
{
  gint wx, wy, bx, by;
  
  gdk_window_get_pointer (text_view->window, &wx, &wy, NULL);
  
  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (text_view), 
                                         GTK_TEXT_WINDOW_WIDGET,
                                         wx, wy, &bx, &by);

  set_cursor_if_appropriate (GTK_TEXT_VIEW (text_view), bx, by);

  return FALSE;
}

GtkWidget * info_init (gchar *selection) {
  GtkWidget     *view;

  view = gtk_text_view_new ();
  info_show_overview (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)), selection); 

  gtk_text_view_set_editable (GTK_TEXT_VIEW(view), FALSE);

  hand_cursor = gdk_cursor_new (GDK_HAND2);
  regular_cursor = gdk_cursor_new (GDK_XTERM);

  g_signal_connect (view, "key-press-event", G_CALLBACK (info_key_pressed), NULL);
  g_signal_connect (view, "event-after", G_CALLBACK (info_event_after), NULL);
  g_signal_connect (view, "motion-notify-event", G_CALLBACK (info_motion_notify_event), NULL);
  g_signal_connect (view, "visibility-notify-event", G_CALLBACK (info_visibility_notify_event), NULL);

  return view;
}

/*
 * TEXT (simple)  --------------------------------------------------------
 *
 */

GtkWidget * text_set (gchar * selection) {
  GtkTextIter    iter;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (tview));
  gtk_text_buffer_set_text (buffer, "", 0);
  gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
  gtk_text_buffer_insert (buffer, &iter, selection, -1);
}

GtkWidget * text_init (gchar * selection) {
  gchar *s;
  GtkTextBuffer *buffer;
  GtkWidget     *view;
  GtkTextIter    iter;

  s=selection?selection:"";

  buffer = gtk_text_buffer_new (NULL);
  gtk_text_buffer_get_start_iter (buffer, &iter);
  gtk_text_buffer_insert (buffer, &iter, s, -1);
  view = gtk_text_view_new_with_buffer (buffer);
  gtk_text_view_set_editable (GTK_TEXT_VIEW(view), FALSE);
  return view;
}

/*
 * MISC  -----------------------------------------------------------------
 *
 */

// clipboard
gchar *get_selection () {
  GtkClipboard *clip;
  clip=gtk_clipboard_get (GDK_SELECTION_PRIMARY);
  return gtk_clipboard_wait_for_text(clip); 
}

/*
 * TOOLBAR  --------------------------------------------------------------
 *
 */

void action_paste (GtkWidget *widget,  gpointer item) {
  const char *name = gtk_widget_get_name(widget);
  GdkAtom sel;
  GtkListStore *model;
  GtkClipboard *clip;
  gchar *s;
  gboolean noclip=FALSE;

  if (!strcmp(name,"paste_clip"))
  {
    sel=GDK_SELECTION_CLIPBOARD; selection_source="CLIPBOARD";
  }
  else if (!strcmp(name,"paste_primary"))
  {
    sel=GDK_SELECTION_PRIMARY; selection_source="PRIMARY Selection";
  }
  else if (!strcmp(name,"paste_second"))
  {
    sel=GDK_SELECTION_SECONDARY; selection_source="SECONDARY Selection";
  }
  else if (!strcmp(name,"clear"))
  {
    noclip=TRUE; selection_source="«none»";
  }

  if (noclip)
    s=NULL;
  else
  {
    clip=gtk_clipboard_get (sel);
    s=gtk_clipboard_wait_for_text(clip);
  }

  // DBG: printf("(%s»[%s]\n", selection_source, s);

  if (!s) s="";
  selection = s;
  // tab 1: list view..
  model = make_model (selection);
  gtk_tree_view_set_model (GTK_TREE_VIEW(list), GTK_TREE_MODEL(model));
  // tab 2: text..
  text_set (selection);
  // tab 3: info..
  info_show_overview (gtk_text_view_get_buffer (GTK_TEXT_VIEW (iview)), selection);

  set_statusbar ();
}

void info_set_to_index (GtkWidget *widget,  gpointer item) {
  info_show_overview (gtk_text_view_get_buffer (GTK_TEXT_VIEW (iview)), selection);
}

GtkToolItem *create_toolbar_button (GtkWidget *toolbar, const gchar *stock_id, const gchar *wid, const gchar *tooltip) {
  GtkToolItem *item;
  item = gtk_tool_button_new_from_stock(stock_id);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), item, -1);
  gtk_widget_set_name(GTK_WIDGET(item), wid);
  if (tooltip) gtk_tool_item_set_tooltip_markup (item, tooltip);
  return item;
}

GtkWidget * create_toolbar () {
  GtkWidget *toolbar;

  GtkToolItem *sep;

  toolbar = gtk_toolbar_new();
  gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
  gtk_container_set_border_width(GTK_CONTAINER(toolbar), 2);
  
  GtkToolItem *paste_primary = create_toolbar_button (toolbar, GTK_STOCK_CONVERT, "paste_primary", "Paste from the <b>Primary Selection</b>");
  g_signal_connect(G_OBJECT(paste_primary), "clicked", G_CALLBACK(action_paste), NULL);
  GtkToolItem *paste_clip = create_toolbar_button (toolbar, GTK_STOCK_PASTE, "paste_clip", "Paste from the <b>Clipboard</b>");
  g_signal_connect(G_OBJECT(paste_clip), "clicked", G_CALLBACK(action_paste), NULL);
  GtkToolItem *paste_second = create_toolbar_button (toolbar, GTK_STOCK_CONVERT, "paste_second", "Paste from the <b>Secondary Selection</b>");
  g_signal_connect(G_OBJECT(paste_second), "clicked", G_CALLBACK(action_paste), NULL);
  GtkToolItem *clear = create_toolbar_button (toolbar, GTK_STOCK_CLEAR, "clear", "Clear...");
  g_signal_connect(G_OBJECT(clear), "clicked", G_CALLBACK(action_paste), NULL);

  sep = gtk_separator_tool_item_new();
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), sep, -1);

  button_index = create_toolbar_button (toolbar, GTK_STOCK_INDEX, "index", "Show <b>Index</b> Info");
  gtk_widget_set_sensitive(GTK_WIDGET(button_index), FALSE);
  g_signal_connect(G_OBJECT(button_index), "clicked", G_CALLBACK(info_set_to_index), NULL);

  toolbar_entry = gtk_entry_new_with_max_length(265); 
  gtk_entry_set_text (GTK_ENTRY(toolbar_entry), "");
  gtk_widget_set_sensitive(GTK_WIDGET(toolbar_entry), FALSE);
  gtk_editable_set_editable(GTK_EDITABLE(toolbar_entry), FALSE);
  GtkToolItem * entry_item = gtk_tool_item_new(); 
  gtk_tool_item_set_expand (GTK_TOOL_ITEM(entry_item), TRUE);
  gtk_container_add (GTK_CONTAINER(entry_item), toolbar_entry);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), entry_item, -1);

  sep = gtk_separator_tool_item_new();
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), sep, -1);

  GtkToolItem *quit = create_toolbar_button (toolbar, GTK_STOCK_QUIT, "quit", "Bye bye!");
  g_signal_connect(G_OBJECT(quit), "clicked", G_CALLBACK(gtk_main_quit), NULL);

  return toolbar;
}

void set_statusbar () {
  gchar *statusline;
  statusline = g_strdup_printf("Source: %s, %d characters.", selection_source, g_utf8_strlen(selection, -1)); 
  gtk_statusbar_pop (GTK_STATUSBAR(statusbar), statusbar_context);
  gtk_statusbar_push (GTK_STATUSBAR(statusbar), statusbar_context, statusline);
}

/*
 * MAIN  -----------------------------------------------------------------
 *
 */

void main_app () {
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *toolbar;
  GtkWidget *swv1, *swv2, *swv3, *swv4;

  // main window ----------------------
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Uninspect: The Unicode Inspector");

  g_signal_connect (window, "destroy", G_CALLBACK (gtk_main_quit), &window);
  g_signal_connect (window, "key-press-event", G_CALLBACK(on_key_press), NULL);

  gtk_container_set_border_width (GTK_CONTAINER (window), 8);

  vbox = gtk_vbox_new (FALSE, 8);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  // == TOOLBAR ===========================================================

  toolbar = create_toolbar();
  gtk_box_pack_start (GTK_BOX (vbox), toolbar, FALSE, FALSE, 0);

  // list -----------------------------====================================
  GtkListStore *model;
  model=make_model(selection=get_selection()); // create listview model from the PRIMARY selection...
  selection_source="PRIMARY Selection";
  list=gtk_tree_view_new_with_model (GTK_TREE_MODEL(model));
  list_init(list);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (list), TRUE);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (list), LVIEW_ID);

  // scrolled window (list inside) ----
  swv1=gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swv1), GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swv1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (swv1), list);

  // text = text view =======================================================
  tview = text_init (selection);
  swv2=gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swv2), GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swv2), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (swv2), tview);

  // info = text view =======================================================
  iview = info_init (selection);
  swv3=gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swv3), GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swv3), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER (swv3), iview);

  // histogram  =============================================================
  // TODO TODO TODO TODO
  swv4=gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swv3), GTK_SHADOW_ETCHED_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swv3), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  // gtk_container_add (GTK_CONTAINER (swv3), iview);

  // notebook -------------------------
  notebook=gtk_notebook_new();
  gtk_widget_set_size_request (notebook, 790, 400);
  gtk_notebook_append_page (GTK_NOTEBOOK(notebook), swv1, gtk_label_new("Codepoints"));
  gtk_notebook_append_page (GTK_NOTEBOOK(notebook), swv3, gtk_label_new("Info"));
  gtk_notebook_append_page (GTK_NOTEBOOK(notebook), swv2, gtk_label_new("Text"));
//  gtk_notebook_append_page (GTK_NOTEBOOK(notebook), swv4, gtk_label_new("Histogram"));
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);

  // STATUS BAR =============================================================
  statusbar = gtk_statusbar_new ();
  statusbar_context = gtk_statusbar_get_context_id (GTK_STATUSBAR(statusbar), "uninspect status");
  set_statusbar ();
  gtk_box_pack_start (GTK_BOX (vbox), statusbar, FALSE, FALSE, 0);

  // window properties
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);

  gtk_widget_show_all (window);
}



int main (int argc, char **argv) {


  g_type_init ();
  gtk_init(NULL, NULL);

  main_app();
  gtk_main();

  return 0; 
}

